{ ****************************************** }
{ Main unit for FPC application "BlackJack2" }
{ ****************************************** }

unit blackjack2_u1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, Menus, ExtCtrls, StdCtrls, PopupNotifier, blackjack2_u2;

{ Types }

type
  TSuit  = (diamond, club, heart, spade);
  TCard  = record
    Suit   : TSuit;
    Rank   : Integer;
    BJValue: Integer;
  end;
  TCardDeck = array[1..52] of TCard;
  TCards = array[1..8] of record
    FaceUp: Boolean;
    Card  : TCard;
  end;
  TImages = array[1..2, 1..8] of TImage;
  // TForm1
  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Game: TMenuItem;
    Game1: TMenuItem;
    Game2: TMenuItem;
    Options: TMenuItem;
    OptionsBJ: TMenuItem;
    OptionsBust: TMenuItem;
    OptionsStandOff: TMenuItem;
    OptionsHold: TMenuItem;
    Help: TMenuItem;
    HelpBJ: TMenuItem;
    HelpProgram: TMenuItem;
    HelpAbout: TMenuItem;
    Label1: TLabel;
    Name1: TEdit;
    Label2: TLabel;
    Name2: TEdit;
    StaticText1: TStaticText;
    Image1, Image2, Image3, Image4, Image5, Image6, Image7, Image8: TImage;
    PlayerNameCards: TStaticText;
    Image9, Image10, Image11, Image12, Image13, Image14, Image15, Image16: TImage;
    StaticText2: TStaticText;
    Label3: TLabel;
    DealerValue: TEdit;
    PlayerName: TLabel;
    PlayerValue: TEdit;
    GameResult: TEdit;
    StaticText3: TStaticText;
    PlayerName1: TLabel;
    Balance1: TEdit;
    PlayerName2: TLabel;
    Balance2: TEdit;
    ButtonDeal: TButton;
    Label4: TLabel;
    Bet: TEdit;
    ButtonHit: TButton;
    ButtonStay: TButton;
    Popup: TPopupNotifier;
    procedure FormCreate(Sender: TObject);
    procedure Game1Click(Sender: TObject);
    procedure Game2Click(Sender: TObject);
    procedure OptionsBJClick(Sender: TObject);
    procedure OptionsBustClick(Sender: TObject);
    procedure OptionsStandOffClick(Sender: TObject);
    procedure OptionsHoldClick(Sender: TObject);
    procedure HelpBJClick(Sender: TObject);
    procedure HelpProgramClick(Sender: TObject);
    procedure HelpAboutClick(Sender: TObject);
    procedure ButtonDealClick(Sender: TObject);
    procedure ButtonHitClick(Sender: TObject);
    procedure ButtonStayClick(Sender: TObject);
    procedure Name1Change(Sender: TObject);
    procedure Name2Change(Sender: TObject);
  end;

{ Global variables }

var
  Form1 : TForm1;
  Images: TImages;
  CardDeck: TCardDeck;
  DealerCards, PlayerCards: TCards;
  DealerBJValue1, PlayerBJValue1, DealerBJValue2, PlayerBJValue2: Integer;
  DealerHold, NCard, BalanceP1, BalanceP2: Integer;
  MultWin, MultBust: Real;
  NameP1, NameP2, ActualPlayer: string;

{ Procedures and functions }

procedure NewGame;
procedure ResetGame;
procedure Shuffle(var CardDeck: TCardDeck);
function NextDeckCard(N: Integer): Integer;
procedure BJValue(Cards: TCards; var V1, V2: Integer);
function GetDealerDoneValue: Integer;
procedure DisplayCard(Player, IX: Integer; Cards: TCards);
function CardFilename(Card: TCard): string;

implementation

{$R *.lfm}

{ New game }

procedure NewGame;

begin
  NameP1 := Form1.Name1.Text; NameP2 := Form1.Name2.Text; ActualPlayer := NameP1;
  Form1.PlayerName.Caption := NameP1; Form1.PlayerNameCards.Caption := NameP1;
  Form1.Name1.Color := clAqua; Form1.Name2.Color := cldefault;
  BalanceP1 := 1000; BalanceP2 := 1000;
  Form1.Balance1.Color := clDefault; Form1.Balance2.Color := clDefault;
  Form1.Balance1.Caption := '1000'; Form1.Balance2.Caption := '1000'; Form1.Bet.Text := '0';
  ResetGame;                                                                   // reset for new round
  Shuffle(CardDeck);                                                           // shuffle the cards
  NCard := 0;
end;

{ Reset the game (for new round) }

procedure ResetGame;

var
  I: Integer;

begin
  DealerBJValue1 := 0; PlayerBJValue1 := 0; DealerBJValue2 := 0; PlayerBJValue2 := 0;
  Form1.DealerValue.Caption := '0'; Form1.PlayerValue.Caption := '0';
  // Remove all cards from the table
  for I := 1 to 8 do begin
    PlayerCards[I].FaceUp := False;
    DealerCards[I].FaceUp := False;
    Images[1, I].Picture.Clear;
    Images[2, I].Picture.Clear;
  end;
  // Reset form controls
  Form1.PlayerName.Caption := ActualPlayer; Form1.PlayerNameCards.Caption := ActualPlayer;
  Form1.GameResult.Caption := '';
  Form1.ButtonDeal.Enabled := True;
  Form1.ButtonHit.Enabled := False;
  Form1.ButtonStay.Enabled := False;
end;

{ Shuffle the cards }

procedure Shuffle(var CardDeck: TCardDeck);

var
  R1, R2, I: Integer;
  Card: TCard;

begin
  for I := 1 TO 50 do begin
    R1 := Random(52) + 1; R2 := Random(52) + 1;
    Card := CardDeck[R1]; CardDeck[R1] := CardDeck[R2]; CardDeck[R2] := Card;
  end;
end;

{ Function: Get next card (index) from deck }

function NextDeckCard(N: Integer): Integer;

begin
  Inc(N);
  if N > 52 then begin                                                         // if all cards have been used, reshuffle
    Shuffle(CardDeck);
    N := 1;
  end;
  NextDeckCard := N;
end;

{ Calculate cards' Blackjack value }

procedure BJValue(Cards: TCards; var V1, V2: Integer);

var
  I: Integer;

begin
  V1 := 0; V2 := 0;
  for I := 1 to 8 do begin
    if Cards[I].FaceUp then begin
      V1 := V1 + Cards[I].Card.BJValue;
      // As may be counted as 1 or as 11
      if Cards[I].Card.Rank = 1 then
        V2 := V2 + 11
      else
        V2 := V2 + Cards[I].Card.BJValue;
    end;
  end;
  // Blackjack (2 aces)
  if (V1 = 2) and (V2 = 22) then begin
    V1 := -21; V2 := -21;                                                      // use -21 to represent Blackjack
  end;
end;

{ Get dealer's hold Blackjack value }

function GetDealerDoneValue: Integer;

var
  DoneValue: Integer;

begin
  DoneValue := 0;
  BJValue(DealerCards, DealerBJValue1, DealerBJValue2);
  if (DealerBJValue1 = -21) and (DealerBJValue2 = -21) then begin              // Blackjack
    Form1.DealerValue.Text := 'BJ';
    DoneValue := -21;
  end
  else begin
    if DealerBJValue2 <= 21 then
      Form1.DealerValue.Text := IntToStr(DealerBJValue2)
    else
      Form1.DealerValue.Text := IntToStr(DealerBJValue1);
    if (DealerBJValue2 >= DealerHold) and (DealerBJValue2 <= 21) then
      DoneValue := DealerBJValue2
    else if (DealerBJValue1 >= DealerHold) and (DealerBJValue1 <= 21) then
      DoneValue := DealerBJValue1
    else if (DealerBJValue1 > 21) and (DealerBJValue2 > 21) then begin
      if DealerBJValue1 < DealerBJValue2 then
        DoneValue := DealerBJValue1
      else
        DoneValue := DealerBJValue2;
    end;
  end;
  GetDealerDoneValue := DoneValue;
end;

{ Display the dealer/player's cards }

procedure DisplayCard(Player, IX: Integer; Cards: TCards);

const
  CardsPath    = 'cmagedeck/';
  DeckCardFile = '155.png';

var
  CardFile: string;

begin
  // Display card face resp. card bottom
  if Cards[IX].FaceUp then
    CardFile := CardFilename(Cards[IX].Card)
  else
    CardFile := DeckCardFile;
  CardFile := CardsPath + CardFile;
  DoDirSeparators(CardFile);                                                   // platform independent file path
  Images[Player, IX].Picture.LoadFromFile(CardFile);
end;

{ Function: Get card image filename }

function CardFilename(Card: TCard): string;

var
  N: Integer;

begin
  N := 100 + Ord(Card.Suit) * 13 + Card.Rank;
  CardFilename := IntToStr(N) + '.png';
end;

{ ----------------- }
{ TForm1 procedures }
{ ----------------- }

{ Initialize the game at application start }

procedure TForm1.FormCreate(Sender: TObject);

var
  Suit: TSuit;
  N, Rank: Integer;

begin
  // Create an array containing the 16 card images
  Images[1, 1] := Image1;  Images[1, 2] := Image2;  Images[1, 3] := Image3;  Images[1, 4] := Image4;
  Images[1, 5] := Image5;  Images[1, 6] := Image6;  Images[1, 7] := Image7;  Images[1, 8] := Image8;
  Images[2, 1] := Image9;  Images[2, 2] := Image10; Images[2, 3] := Image11; Images[2, 4] := Image12;
  Images[2, 5] := Image13; Images[2, 6] := Image14; Images[2, 7] := Image15; Images[2, 8] := Image16;
  // Fill the carddeck
  N := 0;
  for Suit := diamond to spade do begin
    for Rank := 1 to 13 do begin
      Inc(N);
      CardDeck[N].Suit := Suit;
      CardDeck[N].Rank := Rank;
      if Rank <= 10 then
        CardDeck[N].BJValue := Rank
      else
        CardDeck[N].BJValue := 10;
    end;
  end;
  // Default options
  DealerHold := 17; MultWin := 1.5; MultBust := 1;
  Randomize;
  NewGame;                                                                     // start a new game
end;

{ Menu item: "Game > New": start a new game }

procedure TForm1.Game1Click(Sender: TObject);

begin
  NewGame;
end;

{ Menu item "Game > Exit": exit application }

procedure TForm1.Game2Click(Sender: TObject);

begin
  Close;
end;

{ Menu item "Options > Black Jack wins 1.5x the stake": select actual win with Blackjack option }

procedure TForm1.OptionsBJClick(Sender: TObject);

begin
  if OptionsBJ.Checked then begin
    OptionsBJ.Checked := False;
    MultWin := 1;
  end
  else begin
    OptionsBJ.Checked := True;
    MultWin := 1.5;
  end;
end;

{ Menu item "Options > Busting costs 2x the stake": select actual lose by busting option }

procedure TForm1.OptionsBustClick(Sender: TObject);

begin
  if OptionsBust.Checked then begin
    OptionsBust.Checked := False;
    MultBust := 1;
  end
  else begin
    OptionsBust.Checked := True;
    MultBust := 2;
  end;
end;

{ Menu item "Options > Dealer wins at stand-off": select actual stand-off option }

procedure TForm1.OptionsStandOffClick(Sender: TObject);

begin
  if OptionsStandOff.Checked then begin
    OptionsStandOff.Checked := False;
  end
  else begin
    OptionsStandOff.Checked := True;
  end;
end;

{ Menu item "Options > Dealer holds on 18": select actual dealer hold option }

procedure TForm1.OptionsHoldClick(Sender: TObject);

begin
  if OptionsHold.Checked then begin
    OptionsHold.Checked := False;
    DealerHold := 17;
  end
  else begin
    OptionsHold.Checked := True;
    DealerHold := 18;
  end;
end;

{ Menu item "Help > Blackjack": display Blackjack game help }

procedure TForm1.HelpBJClick(Sender: TObject);

begin
  if Form2.Visible then
    Form2.Hide;
  Form2.HelpText.Clear;
  Form2.HelpText.Lines.LoadFromFile('help1.txt');
  Form2.Show;
end;

{ Menu item "Help > Program": display Blackjack 2 program help }

procedure TForm1.HelpProgramClick(Sender: TObject);

begin
  if Form2.Visible then
    Form2.Hide;
  Form2.HelpText.Clear;
  Form2.HelpText.Lines.LoadFromFile('help2.txt');
  Form2.Show;
end;

{ Menu item "Help > About": display Blackjack 2 program info }

procedure TForm1.HelpAboutClick(Sender: TObject);
var
  S: string;

begin
  if Popup.Visible then
    Popup.Visible := False
  else begin
    S := Chr(13) + 'Blackjack game for 2 players' + Chr(13);
    S += 'Version 1.0, © allu, January, 2018';
    Popup.Text := S;
    Popup.Visible := True;
  end;
end;

{ Button "Deal" pressed: Deal the first 4 cards }

procedure TForm1.ButtonDealClick(Sender: TObject);

begin
  ResetGame;
  // Deal first 4 cards
  NCard := NextDeckCard(NCard); PlayerCards[1].Card := CardDeck[NCard];
  NCard := NextDeckCard(NCard); DealerCards[1].Card := CardDeck[NCard];
  NCard := NextDeckCard(NCard); PlayerCards[2].Card := CardDeck[NCard];
  NCard := NextDeckCard(NCard); DealerCards[2].Card := CardDeck[NCard];
  PlayerCards[1].FaceUp := True; DealerCards[1].FaceUp := True;
  PlayerCards[2].FaceUp := True;
  DisplayCard(2, 1, PlayerCards); DisplayCard(1, 1, DealerCards);
  DisplayCard(2, 2, PlayerCards); DisplayCard(1, 2, DealerCards);
  // Calculate Blackjack cards values
  BJValue(DealerCards, DealerBJValue1, DealerBJValue2);
  BJValue(PlayerCards, PlayerBJValue1, PlayerBJValue2);
  if (DealerBJValue1 = -21) and (DealerBJValue2 = -21) then                    // Blackjack
    Form1.DealerValue.Text := 'BJ'
  else begin
    Form1.DealerValue.Text := IntToStr(DealerBJValue1);
    if DealerBJValue2 <> DealerBJValue1 then
      Form1.DealerValue.Text := Form1.DealerValue.Text + ' / ' + IntToStr(DealerBJValue2);
  end;
  if (PlayerBJValue1 = -21) and (PlayerBJValue2 = -21) then                    // Blackjack
    Form1.PlayerValue.Text := 'BJ'
  else begin
    Form1.PlayerValue.Text := IntToStr(PlayerBJValue1);
    if PlayerBJValue2 <> PlayerBJValue1 then
      Form1.PlayerValue.Text := Form1.PlayerValue.Text + ' / ' + IntToStr(PlayerBJValue2);
  end;
  // Enable "Hit" and "Stay" buttons
  Form1.ButtonDeal.Enabled := False;
  Form1.ButtonHit.Enabled := True;
  Form1.ButtonStay.Enabled := True;
  Form1.Bet.SetFocus;
end;

{ Button "Hit" pressed: deal a further card }

procedure TForm1.ButtonHitClick(Sender: TObject);

var
  NextCard, NBet, NBalance, I: Integer;
  Mess: string;
  Colour: TColor;

begin
  if Form1.Bet.Text = '' then
    NBet := 0
  else
    NBet := StrToInt(Form1.Bet.Text);
  if ActualPlayer = NameP1 then
    NBalance := StrToInt(Balance1.Text)
  else
    NBalance := StrToInt(Balance2.Text);
  // If a valid bet has been entered, continue with next player card
  if (NBet >= 10) and (NBet <= NBalance) then begin
    // Get next card from deck and display it
    NCard := NextDeckCard(NCard);
    for I := 1 to 7 do begin
      if PlayerCards[I].FaceUp then
        NextCard := I;
    end;
    Inc(NextCard);
    PlayerCards[NextCard].FaceUp := True;
    PlayerCards[NextCard].Card := CardDeck[NCard];
    DisplayCard(2, NextCard, PlayerCards);
    // Calculate new Blackjack cards value
    BJValue(PlayerCards, PlayerBJValue1, PlayerBJValue2);
    if (PlayerBJValue1 = -21) and (PlayerBJValue2 = -21) then                    // Blackjack
      PlayerValue.Text := 'BJ'
    else begin
      PlayerValue.Text := IntToStr(PlayerBJValue1);
      if (PlayerBJValue2 <> PlayerBJValue1) and (PlayerBJValue2 <= 21) then
        PlayerValue.Text := Form1.PlayerValue.Text + ' / ' + IntToStr(PlayerBJValue2);
    end;
    // Blackjack cards value > 21: player busted and loses
    if PlayerBJValue1 > 21 then begin
      GameResult.Font.Color := clRed;
      GameResult.Text := 'You busted!';
      NBalance -= Trunc(MultBust * NBet);                                     // busting costs 1x or 1.5x the bet (depending on option selected)
      if ActualPlayer = NameP1 then
        BalanceP1 := NBalance
      else
        BalanceP2 := NBalance;
      Balance1.Caption := IntToStr(BalanceP1);
      Balance2.Caption := IntToStr(BalanceP2);
      // No more enough for next bet: end of game
      if NBalance < 10 then begin
        Colour := clyellow;
        if NBalance < 0 then
          Colour := clRed;
        if ActualPlayer = NameP1 then
          Balance1.Color := Colour
        else
          Balance2.Color := Colour;
        MessageDlg('Game over', ActualPlayer + ' is on the rocks!', mtInformation, [mbOK], 0);
      end
      // Blackjack cards value <= 21: game round continues (waiting for another "Hit" or "Stay")
      else begin
        if ActualPlayer = NameP1 then begin
          ActualPlayer := NameP2;
          Name1.Color := clDefault;
          Name2.Color := clAqua;
        end
        else begin
          ActualPlayer := NameP1;
          Name2.Color := clDefault;
          Name1.Color := clAqua;
        end;
        ButtonDeal.Enabled := True;
      end;
      ButtonHit.Enabled := False;
      ButtonStay.Enabled := False;
    end;
  end
  // Invalid bet (minimum of 10 is an arbitrary choice)
  else begin
    if NBet < 10 then
      Mess := 'Bet must be greater than or equal equal to 10!'
    else
      Mess := 'Bet must be less than or equal equal to your balance!';
    MessageDlg('Invalid bet', Mess, mtError, [mbOK], 0);
    Bet.SetFocus;
  end;
end;

{ Button "Stay" pressed: deal dealer's card(s) }

procedure TForm1.ButtonStayClick(Sender: TObject);

var
  DealerDoneValue, PlayerDoneValue, NextCard, NBet, NBalance: Integer;
  Winner, Mess: string;
  Colour: TColor;

begin
  if Form1.Bet.Text = '' then
    NBet := 0
  else
    NBet := StrToInt(Form1.Bet.Text);
  if ActualPlayer = NameP1 then
    NBalance := StrToInt(Balance1.Text)
  else
    NBalance := StrToInt(Balance2.Text);
  // If a valid bet has been entered, continue with next dealer card
  if (NBet >= 10) and (NBet <= NBalance) then begin
    DealerCards[2].FaceUp := True;
    DisplayCard(1, 2, DealerCards);
    NextCard := 3;
    DealerDoneValue := GetDealerDoneValue;
    // If dealer hasn't reached the hold value, continue dealing cards until he'll have
    if DealerDoneValue = 0 then begin
      repeat
        NCard := NextDeckCard(NCard);
        DealerCards[NextCard].FaceUp := True;
        DealerCards[NextCard].Card := CardDeck[NCard];
        DisplayCard(1, NextCard, DealerCards);
        DealerDoneValue := GetDealerDoneValue;
        if DealerDoneValue = 0 then
          Inc(NextCard);
      until DealerDoneValue <> 0;
    end;
    // Find the winner of this round
    if PlayerBJValue2 > 21 then
      PlayerDoneValue := PlayerBJValue1
    else
      PlayerDoneValue := PlayerBJValue2;
    if DealerDoneValue = -21 then begin                                        // dealer has Blackjack (dealer wins or "stand-off")
      if OptionsStandOff.Checked then
        Winner := 'dealer'
      else
        Winner := 'nobody';
    end
    else if PlayerDoneValue = -21 then                                         // player has Blackjack (player wins if dealer hasn't)
      Winner := 'player'
    else begin
      if (DealerDoneValue > 21)                                                // dealer busted (dealer loses)
      or (PlayerDoneValue > DealerDoneValue) then                              // player BJ value > dealer BJ value (player wins)
        Winner := 'player'
      else if DealerDoneValue > PlayerDoneValue then                           // dealer BJ value > player BJ value (dealer wins)
        Winner := 'dealer'
      else if DealerDoneValue = PlayerDoneValue then begin                     // dealer BJ value = player BJ value (dealer wins or "stand-off")
        if OptionsStandOff.Checked then
          Winner := 'dealer'
        else
          Winner := 'nobody';
      end;
    end;
    if Winner = 'player' then begin
      // Round won by player
      GameResult.Font.Color := clDefault;
      GameResult.Text := 'You win!';
      if OptionsBJ.Checked and (PlayerDoneValue = -21) then
        NBalance += Trunc(MultWin * NBet)                                      // Blackjack wins 1.5x the bet (unless unselected)
      else
        NBalance += NBet;
      if ActualPlayer = NameP1 then
        BalanceP1 := NBalance
      else
        BalanceP2 := NBalance;
    end
    else if Winner = 'dealer' then begin
      // Round won by dealer
      GameResult.Font.Color := clDefault;
      GameResult.Text := 'Dealer wins';
      NBalance -= NBet;
      if ActualPlayer = NameP1 then
        BalanceP1 := NBalance
      else
        BalanceP2 := NBalance;
    end
    else begin
      // Stand-off
      GameResult.Font.Color := clDefault;
      GameResult.Text := 'Stand-off';
    end;
    Balance1.Caption := IntToStr(BalanceP1);
    Balance2.Caption := IntToStr(BalanceP2);
    // No more enough for next bet: game over
    if NBalance < 10 then begin
      Colour := clyellow;
      if ActualPlayer = NameP1 then
        Balance1.Color := Colour
      else
        Balance2.Color := Colour;
      MessageDlg('Game over', ActualPlayer + ' is on the rocks!', mtInformation, [mbOK], 0);
    end
    // Continue game with other player now getting cards
    else begin
      if ActualPlayer = NameP1 then begin
        ActualPlayer := NameP2;
        Name1.Color := clDefault;
        Name2.Color := clAqua;
      end
      else begin
        ActualPlayer := NameP1;
        Name2.Color := clDefault;
        Name1.Color := clAqua;
      end;
      ButtonDeal.Enabled := True;
    end;
    ButtonHit.Enabled := False;
    ButtonStay.Enabled := False;
  end
  // Invalid bet
  else begin
    if NBet < 10 then
      Mess := 'Bet must be greater than or equal equal to 10!'
    else
      Mess := 'Bet must be less than or equal equal to your balance!';
    MessageDlg('Invalid bet', Mess, mtError, [mbOK], 0);
    Bet.SetFocus;
  end;
end;

{ Adapt players' names if there is a modification in one of the player name edit fields }

procedure TForm1.Name1Change(Sender: TObject);

var
  OldName: string;

begin
  OldName := NameP1;
  NameP1 := Name1.Text;
  if ActualPlayer = OldName then begin
    ActualPlayer := NameP1;
    PlayerName.Caption := NameP1;
    PlayerNameCards.Caption := NameP1;
  end;
  PlayerName1.Caption := NameP1;
end;

procedure TForm1.Name2Change(Sender: TObject);

var
  OldName: string;

begin
  OldName := NameP2;
  NameP2 := Name2.Text;
  if ActualPlayer = OldName then begin
    ActualPlayer := NameP2;
    PlayerName.Caption := NameP2;
    PlayerNameCards.Caption := NameP2;
  end;
  PlayerName2.Caption := NameP2;
end;

end.

